"""Agent for the vaultgen project."""

from datetime import datetime
from pathlib import Path
from typing import Any

from pydantic_ai import Agent

from .instructions import get_instructions
from .schema import load_schema

MODEL = "google-gla:gemini-3-flash-preview"


def initialize_agent(vault: Path) -> Agent[None, Any]:
    """Initialize the agent for the vaultgen project."""
    schema = load_schema(vault)

    agent = Agent(
        model=MODEL,
        instructions=get_instructions(vault),
        output_type=list[schema] | schema | str,
        output_retries=3,
    )

    @agent.system_prompt
    def _vault_status() -> str:
        parts = []
        if not vault.is_dir():
            return "(There are no notes yet.)"

        for md_path in sorted(vault.rglob("*.md")):
            if md_path.name.lower() == "config.md":
                continue
            try:
                text = md_path.read_text(encoding="utf-8", errors="replace")
                rel = md_path.relative_to(vault)
                parts.append(f"== {rel} ==\n{text}\n")
            except OSError:
                continue

        if len(parts) == 0:
            return "(There are no notes yet.)"

        return "**Current state of the vault:**\n\n" + "\n".join(parts)

    @agent.system_prompt
    def _timestamp() -> str:
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        return f"The current date and time is {now}."

    return agent



"""Exports the notes to the vault."""

from pathlib import Path

from pydantic import BaseModel


def get_note_str(note: BaseModel) -> str:
    """Gets the string representation of a note."""
    header = "---\n"
    for field in note:  # pyright: ignore[reportGeneralTypeIssues]
        name = field[0]
        value = field[1]

        if name == "content" or name == "name":
            continue

        if isinstance(value, str):
            if value.startswith("[[") and value.endswith("]]"):
                value = f"'{value}'"

        if value is None:
            value = ""

        header += f"{name}: {value}\n"

    header += "---\n"

    content = getattr(note, "content", None)

    return header + (content or "")


def export(vault: Path, notes: list[BaseModel] | BaseModel) -> None:
    """Exports the notes to the vault."""
    if not isinstance(notes, list):
        notes = [notes]

    for note in notes:
        note_str = get_note_str(note)

        name = getattr(note, "name", None)
        if name is None:
            raise ValueError("Note name is required")

        note_path = vault.joinpath(f"{name}.md")
        note_path.write_text(note_str)



"""Simple chat CLI that runs a pydantic-ai agent until it returns a structured result."""

from pathlib import Path
from typing import Any

from pydantic_ai import Agent

from .export import export, get_note_str


def cli(agent: Agent[None, Any], vault: Path, hot_start: str | None = None) -> None:
    """Simple chat CLI; exits when the agent returns a non-str (structured result).

    Args:
        agent: The agent to run.
        vault: The vault to use.
        hot_start: The hot start message to use.
    """
    print("\n> Agent: What do you want to generate?")

    message_history: list[Any] = []
    while True:
        if not hot_start:
            user_message = input("\n> You: ").strip()

        else:
            user_message = hot_start
            hot_start = None

        if not user_message:
            continue

        result = agent.run_sync(user_message, message_history=message_history)
        output = result.output
        if isinstance(output, str):
            print(f"\n> Agent: {output}")
            message_history = result.all_messages()

        else:
            print("\n> Agent: ")
            if isinstance(output, list):
                for note in output:
                    print(f"\n== {note.name} ==\n" + get_note_str(note))
            else:
                print(f"\n== {output.name} ==\n" + get_note_str(output))

            user_confirmation = input(
                "\n> Are you okay with this output? (y for yes, or describe what you would like to change): "
            ).strip()

            if user_confirmation == "y":
                export(vault, output)
            else:
                hot_start = user_confirmation



"""Instructions for the vaultgen agent."""

import re
from pathlib import Path


def load_raw_context(vault: Path) -> str:
    """Load context from the ```context code block in config.md (Markdown)."""
    path = vault.joinpath("config.md")
    text = path.read_text()
    match = re.search(r"```context\s*\n(.*?)```", text, re.DOTALL)
    if not match:
        raise ValueError(f"No ```context``` block found in {path}")
    return match.group(1).strip()


def get_instructions(vault: Path) -> str:
    """Build agent instructions including all markdown file contents from the vault."""
    base = (
        "You are a powerful assistant with full access to an Obsidian vault. "
        "You can read, create, and update notes; follow schemas and templates; "
        "and help organize knowledge across the vault. Use the vault contents "
        "below as context—treat it as your working memory. When generating or "
        "editing notes, match the style and structure of existing content and "
        "respect any schema or conventions defined in the vault.\n\n"
        "To create a new note, output the given schema with the content you want; "
        "the note will be generated from it.\n\n"
        "To update a note, output the schema with the same name as the note "
        "to be updated; all content in that note will be overwritten.\n\n"
    )

    context = "**Context of the vault:**\n\n" + load_raw_context(vault) + "\n\n"

    parts = [base, context, "**Current state of the vault:**\n\n"]

    return "".join(parts)



"""Start the vaultgen project."""

from pathlib import Path

CONTENT = """## Context

```context
Here you can define the context of the vault.
```


## Schema

```json
{
    "property": {
        "type": "str" | "int" | "float" | "bool" | "date" | "datetime" | "list" | "link" | ["Option1", "Option2", "Option3"],
        "description": "A property of the note"
        "optional": "True" | "False" // This field is optional if the property is not required.
    }
}
```
"""


def start(vault: Path) -> None:
    """Start the vaultgen project."""
    path = vault.joinpath("config.md")
    path.write_text(CONTENT)



"""Main entry point for the vaultgen project."""

import argparse
from pathlib import Path

from .agent import initialize_agent
from .cli import cli
from .start import start


def chat() -> None:
    """Chat function for the vaultgen project."""
    vault = Path.cwd()
    if not vault.exists():
        raise FileNotFoundError(f"Vault not found in {vault}")

    agent = initialize_agent(vault)
    cli(agent, vault)


def init() -> None:
    """Init function for the vaultgen project."""
    vault = Path.cwd()
    if not vault.exists():
        raise FileNotFoundError(f"Vault not found in {vault}")

    start(vault)


def main() -> None:
    """Main function for the vaultgen project."""
    parser = argparse.ArgumentParser(description="Vaultgen project")
    subparsers = parser.add_subparsers(dest="command", required=True)

    subparsers.add_parser(
        "init", help="Initialize the vaultgen project (creates config.md)"
    )
    subparsers.add_parser("chat", help="Chat with the vaultgen agent")

    args = parser.parse_args()
    if args.command in ("init"):
        init()
    elif args.command == "chat":
        chat()



"""Loads the schema for the vaultgen agent."""

import json
import re
from datetime import date, datetime
from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel, Field, create_model


def load_raw_schema(vault: Path) -> dict:
    """Load schema from the ```json code block in config.md (JSON)."""
    path = vault.joinpath("config.md")
    text = path.read_text()
    match = re.search(r"```json\s*\n(.*?)```", text, re.DOTALL)
    if not match:
        raise ValueError(f"No ```json``` block found in {path}")
    return json.loads(match.group(1).strip())


def get_type(type_str: str) -> Any:
    """Get the type from the string."""
    if isinstance(type_str, list):
        return Literal[*type_str]

    type_map: dict[str, type] = {
        "datetime": datetime,
        "date": date,
        "link": str,
        "list": list,
        "str": str,
        "int": int,
        "float": float,
        "bool": bool,
    }

    if type_str in type_map:
        return type_map[type_str]

    raise ValueError(f"Unknown type: {type_str}")


def load_schema(vault: Path) -> type[BaseModel]:
    """Load schema from the ```json code block in config.md (JSON)."""
    schema = load_raw_schema(vault)
    fields = {}
    for key, value in schema.items():
        field_type = get_type(value["type"])
        if value.get("optional", False):
            field_type = field_type | None

        description = value["description"]
        if value["type"] == "link":
            description += " (it should be the exact name of the linked note in double brackets: e.g. '[[LINKED NOTE NAME]]')"

        fields[key] = (
            field_type,
            Field(description=description),
        )

    fields["name"] = (
        str,
        Field(description="The name of the file and title of the note"),
    )
    fields["content"] = (
        str | None,
        Field(description="The content of the note", default=None),
    )

    docstring = "The schema for the vaultgen agent."
    return create_model("Schema", **fields, __doc__=docstring)
